#include <Ntifs.h>
#include <Ntddk.h>

#include "IntegrityLevel.h"

#ifdef _DEBUG
	#define LOG DbgPrint
#else
	#define LOG(...)
#endif

#define MAX_PATH 260
#define DEV_NAME L"ILDriver"
#define SYMLINK_NAME L"\\??\\" DEV_NAME
#define IOCTL_SET_PROCESS_NAME CTL_CODE(FILE_DEVICE_UNKNOWN, 0x801, METHOD_BUFFERED, FILE_ANY_ACCESS)

typedef struct _ILDeviceExt
{
	PDEVICE_OBJECT DeviceObject;
	UNICODE_STRING SymLink;
} IL_DEVICE_EXT, *PIL_DEVICE_EXT;

#pragma code_seg("INIT")

VOID DriverUnload(PDRIVER_OBJECT pDriverObject);
NTSTATUS DispatchRoutine(PDEVICE_OBJECT DeviceObject, PIRP Irp);


NTSTATUS DriverEntry(PDRIVER_OBJECT DriverObject, PUNICODE_STRING RegistryPath)
{
	NTSTATUS Status;
	PDEVICE_OBJECT DeviceObject;
	UNICODE_STRING DeviceName;
	PIL_DEVICE_EXT DeviceExtension;
	UNICODE_STRING SymLink;
	UINT16 i;


	LOG("DriverEntry");

	Status = STATUS_SUCCESS;
	DeviceObject = NULL;
	SymLink.Length = 0;

	for (;;)
	{
		if (!NT_SUCCESS(Status))
		{
			LOG("DriverEntry - something went wrong");

			if (DeviceObject != NULL)
				IoDeleteDevice(DeviceObject);

			if (SymLink.Length != 0)
				IoDeleteSymbolicLink(&SymLink);

			break;
		}

		for (i = 0; i < IRP_MJ_MAXIMUM_FUNCTION; i++)
			DriverObject->MajorFunction[i] = DispatchRoutine;

		DriverObject->MajorFunction[i] = NULL;
		DriverObject->DriverUnload = DriverUnload;

		RtlInitUnicodeString(&DeviceName, L"\\Device\\" DEV_NAME);

		Status = IoCreateDevice
		(
			DriverObject,
			sizeof(IL_DEVICE_EXT),
			&DeviceName,
			FILE_DEVICE_UNKNOWN,
			0,
			FALSE,
			&DeviceObject
		);

		if (!NT_SUCCESS(Status))
			continue;

		DeviceExtension = (PIL_DEVICE_EXT)DeviceObject->DeviceExtension;
		DeviceExtension->DeviceObject = DeviceObject;

		LOG("DriverEntry - DeviceObject=%p, DeviceExtension=%p", DeviceObject, DeviceExtension);

		RtlInitUnicodeString(&SymLink, SYMLINK_NAME);
		DeviceExtension->SymLink = SymLink;

		Status = IoCreateSymbolicLink(&SymLink, &DeviceName);
		if (!NT_SUCCESS(Status))
			continue;

		Status = DeployProcessHook();
		if (!NT_SUCCESS(Status))
			continue;

		break;
	}

	LOG("DriverEntry - status=%08X", Status);

	return Status;
}

#pragma code_seg()

NTSTATUS DispatchRoutine(PDEVICE_OBJECT DeviceObject, PIRP Irp)
{
	PIL_DEVICE_EXT DeviceExtension;
	PIO_STACK_LOCATION pIrpStack;
	BOOLEAN IoResult;
	ULONG outSize, inSize;
	WCHAR ApplicationNameRaw[MAX_PATH];
	UNICODE_STRING ApplicationNameUnicode;

	LOG("DispatchRoutine");
	
	pIrpStack = IoGetCurrentIrpStackLocation(Irp);
	Irp->IoStatus.Information = 0;
	Irp->IoStatus.Status = STATUS_SUCCESS;

	DeviceExtension = (PIL_DEVICE_EXT)DeviceObject->DeviceExtension;

	switch (pIrpStack->MajorFunction)
	{
		case IRP_MJ_CREATE:
			LOG("IRP_MJ_CREATE");
		break;

		case IRP_MJ_CLOSE:
			LOG("IRP_MJ_CLOSE");
		break;

		case IRP_MJ_CLEANUP:
			LOG("IRP_MJ_CLEANUP");
		break;

		case IRP_MJ_DEVICE_CONTROL:
			LOG("IRP_MJ_DEVICE_CONTROL");

			switch (pIrpStack->Parameters.DeviceIoControl.IoControlCode)
			{
				case IOCTL_SET_PROCESS_NAME:
				{
					LOG("IOCTL_SET_PROCESS_NAME");

					outSize = pIrpStack->Parameters.DeviceIoControl.OutputBufferLength;
					inSize = pIrpStack->Parameters.DeviceIoControl.InputBufferLength;

					if (outSize != sizeof(BOOLEAN) || inSize != MAX_PATH * sizeof(WCHAR))
					{
						Irp->IoStatus.Status = STATUS_INVALID_PARAMETER;
						break;
					}

					RtlCopyMemory(ApplicationNameRaw, Irp->AssociatedIrp.SystemBuffer, MAX_PATH);
					ApplicationNameRaw[MAX_PATH - 1] = L'\0';

					RtlInitUnicodeString(&ApplicationNameUnicode, ApplicationNameRaw);
					LOG("UnicodeName=%wZ", &ApplicationNameUnicode);
						

					IoResult = SetTargetProcessName(ApplicationNameUnicode);

					RtlCopyMemory(Irp->AssociatedIrp.SystemBuffer, &IoResult, sizeof(BOOLEAN));
					Irp->IoStatus.Information = sizeof(BOOLEAN);
				}

				break;
			}
		break;

		default:
			LOG("Not implemented major function call - %02X", pIrpStack->MajorFunction);
			Irp->IoStatus.Status = STATUS_NOT_IMPLEMENTED;
		break;
	}

	IoCompleteRequest(Irp, IO_NO_INCREMENT);

	return Irp->IoStatus.Status;
}

VOID DriverUnload(PDRIVER_OBJECT pDriverObject)
{
	PDEVICE_OBJECT NextDevObj;
	NTSTATUS Status;


	LOG("DriverUnload");

	for (NextDevObj = pDriverObject->DeviceObject; NextDevObj != NULL; NextDevObj = NextDevObj->NextDevice)
	{
		PIL_DEVICE_EXT DeviceExtension = (PIL_DEVICE_EXT)NextDevObj->DeviceExtension;
		UNICODE_STRING * Link = &(DeviceExtension->SymLink);

		Status = IoDeleteSymbolicLink(Link);
		if (!NT_SUCCESS(Status))
			LOG("DriverUnload - IoDeleteSymbolicLink=%p, Status=%08X", Link, Status);

		IoDeleteDevice(DeviceExtension->DeviceObject);
	}

	(void)RemoveProcessHook();
}
