#include <Ntifs.h>
#include <Ntddk.h>
#include <ntstrsafe.h>

#ifdef _DEBUG
	#define LOG DbgPrint
#else
	#define LOG(...)
#endif

/* Из winnt.h */
#define SE_GROUP_INTEGRITY (0x00000020L)

static UNICODE_STRING ProcessName;
static KMUTEX NameMutex;

VOID HookRoutine(PEPROCESS Process, HANDLE ProcessId, PPS_CREATE_NOTIFY_INFO CreateInfo);
VOID SetLowIntegrity(PEPROCESS Process);
VOID GetFileName(PUNICODE_STRING pSourceString, PUNICODE_STRING pFileName);
NTSTATUS TuneTokenIntegrity(HANDLE Token);


NTSTATUS DeployProcessHook()
{
	NTSTATUS Status = STATUS_UNSUCCESSFUL;


	LOG("DeployProcessHook");

	for (;;)
	{
		KeInitializeMutex(&NameMutex, 0);

		Status = PsSetCreateProcessNotifyRoutineEx(HookRoutine, FALSE);
		if (!NT_SUCCESS(Status))
			break;

		RtlInitUnicodeString(&ProcessName, NULL);

		break;
	}

	LOG("DeployProcessHook - Status=%08X", Status);

	return Status;
}

NTSTATUS RemoveProcessHook()
{
	NTSTATUS Status = STATUS_UNSUCCESSFUL;

	
	LOG("RemoveProcessHook");

	for (;;)
	{
		Status = PsSetCreateProcessNotifyRoutineEx(HookRoutine, TRUE);
		if (!NT_SUCCESS(Status))
			break;
		
		KeWaitForMutexObject(&NameMutex, Executive, KernelMode, FALSE, NULL);
		RtlFreeUnicodeString(&ProcessName);
		KeReleaseMutex(&NameMutex, FALSE);

		break;
	}

	LOG("RemoveProcessHook - Status=%08X", Status);

	return Status;
}

VOID HookRoutine(PEPROCESS Process, HANDLE ProcessId, PPS_CREATE_NOTIFY_INFO CreateInfo)
{
	NTSTATUS Status;
	LARGE_INTEGER Timeout;
	UNICODE_STRING TargetProcessName, FullPathName, FileName;


	LOG("HookRoutine");

	/* Таймаут на 500 ms */
	Timeout.QuadPart = 5000000L;
	Timeout.QuadPart = -(Timeout.QuadPart);

	for (;;)
	{
		/* Игнорируем следующие ситуации: процесс завершает работу, у процесса отсутствует имя, не задано имя для применения фильтра */
		if (CreateInfo == NULL || !CreateInfo->FileOpenNameAvailable || ProcessName.Length == 0)
			break;

		/* Делаем локальную копию глобальной переменной, чтобы минимизировать время нахождения под мьютексом */
		Status = KeWaitForMutexObject(&NameMutex, UserRequest, KernelMode, FALSE, &Timeout);
		if (Status != STATUS_SUCCESS)
		{
			LOG("KeWaitForMutexObject - Status=%08X", Status);
			break;
		}

		RtlCreateUnicodeString(&TargetProcessName, ProcessName.Buffer);

		KeReleaseMutex(&NameMutex, FALSE);

		/* Получаем имя процесса из полного пути и понижаем уровень целостности если это нужный процесс */
		RtlCreateUnicodeString(&FullPathName, CreateInfo->ImageFileName->Buffer);

		GetFileName(&FullPathName, &FileName);

		LOG("TargetProcessName=%wZ, FullPathName=%wZ, FileName=%wZ", &ProcessName, &FullPathName, &FileName);

		if (RtlCompareUnicodeString(&ProcessName, &FileName, FALSE) == 0)
			SetLowIntegrity(Process);

		RtlFreeUnicodeString(&TargetProcessName);
		RtlFreeUnicodeString(&FullPathName);

		break;
	}
}

BOOLEAN SetTargetProcessName(UNICODE_STRING Name)
{
	BOOLEAN Status = TRUE;


	LOG("SetTargetProcessName");

	KeWaitForMutexObject(&NameMutex, UserRequest, KernelMode, FALSE, NULL);

	for (;;)
	{
		RtlFreeUnicodeString(&ProcessName);

		if(!RtlCreateUnicodeString(&ProcessName, Name.Buffer))
		{
			Status = FALSE;
			LOG("SetTargetProcessName - RtlCreateUnicodeString failed");
			break;
		}
		
		break;
	}

	KeReleaseMutex(&NameMutex, FALSE);

	return Status;
}

VOID SetLowIntegrity(PEPROCESS Process)
{
	PACCESS_TOKEN pAccessToken;
	NTSTATUS Status;
	HANDLE hToken = NULL;


	LOG("SetLowIntegrity");

	Status = STATUS_SUCCESS;
	for (;;)
	{
		if (!NT_SUCCESS(Status))
		{
			LOG("SetLowIntegrity - something went wrong");

			break;
		}

		pAccessToken = PsReferencePrimaryToken(Process);

		Status = ObOpenObjectByPointer(pAccessToken, 0, NULL, TOKEN_QUERY, NULL, KernelMode, &hToken);
		if (!NT_SUCCESS(Status))
			continue;

		Status = TuneTokenIntegrity(hToken);
		if (!NT_SUCCESS(Status))
			continue;

		break;
	}

	if (hToken != NULL)
		ZwClose(hToken);

	PsDereferencePrimaryToken(pAccessToken);

	LOG("SetLowIntegrity - Status=%08X", Status);
}

NTSTATUS TuneTokenIntegrity(HANDLE Token)
{
	PSID lowMandatoryLevelSid;
	TOKEN_MANDATORY_LABEL mandatoryLabel;
	static SID_IDENTIFIER_AUTHORITY mandatoryLabelAuthority = SECURITY_MANDATORY_LABEL_AUTHORITY;
	UCHAR lowMandatoryLevelSidBuffer[FIELD_OFFSET(SID, SubAuthority) + sizeof(ULONG)];


	lowMandatoryLevelSid = (PSID)lowMandatoryLevelSidBuffer;
	RtlInitializeSid(lowMandatoryLevelSid, &mandatoryLabelAuthority, 1);
	*RtlSubAuthoritySid(lowMandatoryLevelSid, 0) = SECURITY_MANDATORY_LOW_RID;

	mandatoryLabel.Label.Sid = lowMandatoryLevelSid;
	mandatoryLabel.Label.Attributes = SE_GROUP_INTEGRITY;

	return ZwSetInformationToken(Token, TokenIntegrityLevel, &mandatoryLabel, sizeof(TOKEN_MANDATORY_LABEL));
}

VOID GetFileName(PUNICODE_STRING pSourceString, PUNICODE_STRING pFileName)
{
	UNICODE_STRING current = *pSourceString;
	UNICODE_STRING remaining;

	for (;;)
	{
		FsRtlDissectName(current, pFileName, &remaining);
		if (remaining.Length == 0)
			break;

		current = remaining;
	}
}
